import os
import re
import traceback
from glob import glob


def path_list(paths: list[str]) -> str:
    return ' '.join(p.replace(' ', '$ ') for p in paths)


class Rule:
    """
    Represents a rule in the ninja build files.

    The primary properties of a rule are the command to execute and a description. Both the command
    and the description can contain variables, whose values will later be set by each `BuildEdge`
    that references the rule. The description is displayed during the build for information
    purposes.
    """

    # pylint: disable=R0917
    def __init__(self, cmd: str, desc: str, deps: str = '', depfile: str = '',
                 generator: str = '', pool: str = '', restat: bool = False):
        """
        Creates a new rule

        Parameters
        ----------
        :param cmd: the command to execute
        :param desc: the description for the command (for information purposes)
        :param deps: either empty, 'gcc' or 'msvc'. The latter specify special dependency processing
            (header files).
        :param depfile: if not empty, a file in Makefile syntax that contains extra dependencies.
            This can be used for the file gcc generates with the `-MD` argument.
        :param generator: if not empty, specifies that this rule is used to re-invoke the generator
            program. See [Ninja manual](https://ninja-build.org/manual.html#ref_rule) for more
            detail.
        :param pool: if not empty, the name of the pool to use. See [Ninja
            manual](https://ninja-build.org/manual.html#ref_pool).
        :param restat: if `True`, causes Ninja to re-stat the command's outputs after execution of
            the command. Each output whose modification time the command did not change will be
            treated as though it had never needed to be built. This may cause the output's reverse
            dependencies to be removed from the list of pending build actions.
        """

        assert cmd != '' and desc != ''
        self.cmd = cmd
        self.desc = desc
        self.deps = deps
        self.depfile = depfile
        self.generator = generator
        self.pool = pool
        self.restat = restat
        self.refs = 0

    def _write_to_file(self, name: str, file):
        """
        Writes this rule to the given file object

        Parameters
        ----------
        :param name: the name of the rule
        :param file: the file object
        """

        file.write('rule %s\n' % name)
        file.write('  command = %s\n' % self.cmd)
        file.write('  description = %s\n' % self.desc)
        if self.deps != '':
            file.write('  deps = %s\n' % self.deps)
        if self.depfile != '':
            file.write('  depfile = %s\n' % self.depfile)
        if self.generator != '':
            file.write('  generator = %s\n' % self.generator)
        if self.pool != '':
            file.write('  pool = %s\n' % self.pool)
        if self.restat:
            file.write('  restat = 1\n')


class BuildEdge:
    """
    A build edge specifies that given output files are produced from given input files using a
    specific rule.

    A build edge therefore references a specific rule and assigns values to the variables defined
    for the rule. Most importantly, a rule has one or more output files and one or more input files.
    Whenever one of the input files changes, the command of the rule will be invoked to reproduce
    the output files. Additional dependencies can be specified to also trigger a rebuild.
    """

    # pylint: disable=R0917
    def __init__(self, rule: str, outs: list[str], ins: list[str], deps: list[str] = None,
                 vars: dict[str, str] = None, libs: list[str] = None, lib_path: list[str] = None):
        """
        Creates a new build edge.

        Parameters
        ----------
        :param rule: the name of the rule (needs to exist in the `Generator`)
        :param outs: a list of paths that are produced as outputs when executing the command
        :param ins: a list of paths that are taken as inputs by the command
        :param deps: an optional list of paths with additional dependencies
        :param vars: an optional list with values for additional variables used in the rule
        :param libs: when producing executables, a list of library names that is linked against
        :param lib_path: when producing executables, a list of paths to search the libraries in
        """

        assert len(outs) > 0, "The list of output files cannot be empty"

        deps = [] if deps is None else deps
        vars = {} if vars is None else vars
        libs = [] if libs is None else libs
        lib_path = [] if lib_path is None else lib_path

        self.calltrace = None
        self.rule = rule
        self.outs = outs
        self.ins = ins
        # copy the dependencies, because we want to alter them for this specific BuildEdge later
        self.deps = deps.copy()
        # without inputs and dependencies, we always want to rebuild it
        if self.rule != 'phony' and len(self.ins) == 0 and len(self.deps) == 0:
            self.deps = ['always']
        self.libs = libs
        self.lib_path = lib_path
        self.vars = vars

    def _write_to_file(self, defaults: dict[str, str], file):
        """
        Writes this build edge to the given file object

        Parameters
        ----------
        :param defaults: a dictionary with default values. Only if a variable has not the default
            value, it will be specified for the build edge.
        :param file: the file object
        """

        file.write('build %s: %s %s' %
                   (path_list(self.outs), self.rule, path_list(self.ins)))
        if len(self.deps) > 0:
            file.write(' | %s' % path_list(self.deps))
        file.write('\n')
        for key, val in self.vars.items():
            if key not in defaults or defaults[key] != val:
                file.write('  %s = %s\n' % (key, val))


class Generator:
    """
    The `Generator` collects rules and build edges and finally writes a Ninja build file.

    The rules describe how specific file types are built (e.g., shared libraries), while a build
    edge describes one concrete file to build. The generator comes with a number of builtin rules
    for C, C++, and Rust, but can be extended by additional rules via `Generator.add_rule`. Build
    edges can be added with `Generator.add_build`.

    Typically, the build edges are added to the generator via the environment (`Env`). For example,
    the method `Env.cc` will add a build edge for the rule `cc` to the generator that is given to
    `Env.cc`. The reasoning is that `Env` holds the required values in terms of tools, compiler
    flags, and paths for the build of a specific file and hence will create a build edge based on
    this information and add them to the `Generator`.
    """

    def __init__(self):
        """
        Creates a new `Generator` with the default rules.
        """

        self._build_dir = os.environ.get('NPBUILD')
        self._debug = os.environ.get('NPDEBUG', '0') == '1'
        self._rules = {}
        self._build_edges = []
        self._globs = []
        self._build_files = []

        # default rules
        self.add_rule('install', Rule(
            cmd='install $instflags $in $out && touch $out',
            desc='INSTALL $out'
        ))
        self.add_rule('cpp', Rule(
            cmd='$cpp -MD -MF $out.d -P $cppflags $in -o $out',
            deps='gcc',
            depfile='$out.d',
            desc='CPP $out'
        ))
        self.add_rule('cc', Rule(
            cmd='$cc -MD -MF $out.d $ccflags -c $in -o $out',
            deps='gcc',
            depfile='$out.d',
            desc='CC $out'
        ))
        self.add_rule('cxx', Rule(
            cmd='$cxx -MD -MF $out.d $cxxflags -c $in -o $out',
            deps='gcc',
            depfile='$out.d',
            desc='CXX $out'
        ))
        self.add_rule('ar', Rule(
            cmd='$ar rc $arflags $out $in && $ranlib $out',
            desc='AR $out'
        ))
        self.add_rule('link', Rule(
            cmd='$link -o $out $in $linkflags',
            desc='LINK $out'
        ))
        self.add_rule('shlink', Rule(
            cmd='$shlink -shared -o $out $in $shlinkflags',
            desc='SHLINK $out'
        ))
        self.add_rule('strip', Rule(
            cmd='$strip -o $out $in',
            desc='STRIP $out'
        ))
        self.add_rule('cargo', Rule(
            cmd='cd $dir && $env $cargo $cargoflags',
            desc='CARGO $out',
            # recheck which output files have changed after the command to only relink the
            # executables where the library generated by Rust actually changed
            restat=True,
        ))

        # special build edge for 'always-rebuild' build edges
        self._build_edges.append(BuildEdge(
            'phony',
            outs=['always'],
            ins=[],
            deps=[]
        ))

    def add_rule(self, name: str, rule: Rule):
        """
        Adds a new rule to this generator.

        These rules can later be referenced when adding build edges via `Generator.add_build`. Note
        that Rule names are unique and cannot be added twice.

        Parameters
        ----------
        :param name: The name of the rule
        :param rule: The `Rule` object to add
        """

        assert name not in self._rules
        self._rules[name] = rule

    def add_build(self, edge: BuildEdge):
        """
        Adds a new build edge to this generator.

        This build edge assigns concrete values to the variables defined by the rule. As the
        referenced rule is now used, the generator will write both the rule and the build edge to
        the Ninja build file when `Generator.write_to_file` is called. Note that the rule this build
        edge refers to needs to exist.

        Parameters
        ----------
        :param name: The name of the rule
        :param rule: The `Rule` object to add
        """

        assert edge.rule in self._rules

        if self._debug:
            for ex_edge in self._build_edges:
                # skip edges that don't have a calltrace (e.g., the always-rebuild rule)
                if ex_edge.calltrace is None:
                    continue

                for out in ex_edge.outs:
                    assert out not in edge.outs, \
                        "Output '{}' is already produced by the build edge added here:\n{}".format(
                            out, ''.join(traceback.format_list(ex_edge.calltrace)))
            edge.calltrace = traceback.extract_stack()

        self._rules[edge.rule].refs += 1
        self._build_edges.append(edge)

    def _add_glob(self, pattern):
        """
        Adds the given pattern to the list of globs

        These globs will be written to file in `Generator.write_to_file`.
        """

        self._globs += [pattern]

    def _add_build_file(self, file):
        """
        Adds the given file to the list of encountered `build.py` files

        This list will be written to file in `Generator.write_to_file`.
        """

        self._build_files += [file]

    def write_to_file(self, defaults: dict[str, str] = None):
        """
        Writes the Ninja build file according to the so far added rules and build edges.

        The written file is stored in `$NPBUILD/build.ninja` and will be overwritten, if it already
        exists. Additionally, a `$NPBUILD/.build.deps` file will be written, which holds the
        dependencies of the regeneration rule. This regeneration rule defines when the build.ninja
        file needs to be regenerated and how. The rule therefore depends on all `build.py` files in
        the current directory or any subdirectory.

        Parameters
        ----------
        :param defaults: the default variables. If the value of a variable for a particular build
            edge is the same as the value in the default variables, the variable will not be
            specified for the build edge. As such, default variables help to reduce the size of the
            ninja build file. By default (if `defaults` is `None`), Ninjapie determines these
            defaults automatically by chosing the most used value for each default variable. You can
            however specify a custom set of default variables instead. For example, this can be used
            for debugging by specifying an empty dict so that all variables will be specified
            explicitly for every build edge, which makes it easier to read the ninja build file.
        """

        outdir = self._build_dir

        build_file = outdir + '/build.ninja'
        dep_file = outdir + '/.build.deps'
        glob_file = outdir + '/.build.globs'

        # rules and build edge to automatically regenerate the build.ninja
        self.add_rule('generator', Rule(
            cmd='python3 -B build.py',
            depfile=outdir + '/.build.deps',
            pool='build_pool',
            generator='1',
            desc='Regenerating build.ninja',
        ))
        this_dir = os.path.dirname(os.path.abspath(__file__))
        self.add_build(BuildEdge(
            'generator',
            outs=[build_file],
            ins=[],
            deps=glob(this_dir + "/*.py"),
        ))

        self._finalize_deps()
        if defaults is None:
            defaults = self._determine_defaults()

        # generate build.ninja
        with open(build_file, 'w', encoding='utf-8') as file:
            file.write('# This file has been generated by the ninjapie build system.\n')
            file.write('\n')

            for key, val in defaults.items():
                file.write('%s = %s\n' % (key, val))
            file.write('\n')

            for name, rule in self._rules.items():
                if rule.refs > 0:
                    rule._write_to_file(name, file)
            file.write('\n')

            # separate pool for the build.ninja regeneration to run that alone
            file.write('pool build_pool\n')
            file.write('  depth = 1\n')
            file.write('\n')

            for edge in self._build_edges:
                edge._write_to_file(defaults, file)

        # generate deps of build.ninja
        build_files = ['build.py'] + self._build_files
        with open(dep_file, 'w', encoding='utf-8') as file:
            file.write(build_file + ': ' + path_list(build_files))

        # generate files with all globs
        with open(glob_file, 'w', encoding='utf-8') as file:
            for glb in self._globs:
                file.write(glb + '\n')

    def write_compile_cmds(self, outdir: str = None):
        """
        Writes a `compiler_commands.json` file for `clangd`.

        This file is leveraged by the language server `clangd` to know how your source files are
        build. The file is written to `$NPBUILD/compile_commands.json` and will be overwritten, if
        existing. Note that only the rules `cxx` and `cc` are considered.
        """

        if outdir is None:
            outdir = self._build_dir

        # generate compile_commands.json for clangd
        with open(outdir + '/compile_commands.json', 'w', encoding='utf-8') as cmds:
            cmds.write('[\n')
            base_dir = os.getcwd()
            count = 0
            for edge in self._build_edges:
                if edge.rule in ('cxx', 'cc'):
                    assert len(edge.ins) == 1
                    if count > 0:
                        cmds.write(',\n')
                    cmds.write('  {\n')
                    cmds.write('    "directory": "{}",\n'.format(base_dir))
                    cmds.write('    "file": "{}",\n'.format(edge.ins[0]))
                    cmds.write('    "command": "{}"\n'.format(self._get_clang_flags(edge)))
                    cmds.write('  }')
                    count += 1
            cmds.write('\n]\n')

    def _get_clang_flags(self, bedge: BuildEdge) -> str:
        """
        Determines the flags for clang or clang++ for the given build edge.

        Parameters
        ----------
        :param bedge: the build edge for which to generate the flags

        Returns
        -------
        A string with the flags
        """

        flags = 'ccflags' if bedge.rule == 'cc' else 'cxxflags'
        flag_str = 'clang' if bedge.rule == 'cc' else 'clang++'
        flag_str += ' ' + bedge.vars[flags].replace('"', '\\"')
        # remove all machine specific flags, because clang does not support all ISAs, etc.
        return re.sub(r'\s+-m\S+', '', flag_str)

    def _determine_defaults(self) -> dict[str, str]:
        """
        Determines the default values for the used variables.

        Ninja supports global variables whose values are used whenever a variable is not set for a
        particular build edge. For that reason, Ninjapie auto-generates these global variables based
        on the most used values in all build edges. This approach thus reduces the size of the
        `build.ninja` file.

        Returns
        -------
        A dictionary with the variable names and values
        """

        # first count the number of times for each value and each variable
        vars = {}
        for edge in self._build_edges:
            for key, val in edge.vars.items():
                if key not in vars:
                    vars[key] = {}
                if val in vars[key]:
                    vars[key][val] += 1
                else:
                    vars[key][val] = 1

        # now use the most-used value for each variable as the default
        defaults = {}
        for name, vals in vars.items():
            max_count = 0
            max_key = None
            for key, count in vals.items():
                if count > max_count:
                    max_key = key
                    max_count = count
            if max_key is not None:
                defaults[name] = max_key
        return defaults

    def _finalize_deps(self):
        """
        Finalizes the dependencies of all build edges.

        When build edges use libraries, we first collect them under the `libs` field. After all
        libraries that we build ourself are known, we complete the dependencies of the build edges
        based on these `libs`. That is, whenever we build the library ourself, we add the path to
        the built library to the dependency. Otherwise, we assume that the library exists in some
        system directory and will not change (we do not add it to the dependencies).
        """

        libs = self._collect_libs()
        for edge in self._build_edges:
            for lib in edge.libs:
                stname = 'lib' + lib + '.a'
                shname = 'lib' + lib + '.so'
                for path in edge.lib_path:
                    if path in libs:
                        if shname in libs[path]:
                            edge.deps.append(libs[path][shname])
                            break
                        if stname in libs[path]:
                            edge.deps.append(libs[path][stname])
                            break

    def _collect_libs(self) -> dict[str, dict[str, str]]:
        """
        Collects the libraries that we build ourself.
        """

        libs = {}
        for edge in self._build_edges:
            for out in edge.outs:
                if out.endswith('.a') or out.endswith('.so'):
                    dir = os.path.dirname(out)
                    if dir not in libs:
                        libs[dir] = {}
                    libs[dir][os.path.basename(out)] = out
        return libs



import os


class SourcePath(str):
    """
    A path for source files

    This class is used to refer to source files. These are specified relative to the current
    directory in the given environment (`Env.cur_dir`).

    `SourcePath` implements `__str__` and `__repr__` and can thus be used as a string (the contained
    path).
    """

    def __init__(self, path: str):
        """
        Creates a new `SourcePath` that simply uses the given path

        Parameters
        ----------
        :param path: the path
        """

        self._path = path

    @staticmethod
    def new(env, path):
        """
        Creates a new `SourcePath` from given path object

        The path object can be a `SourcePath`, `BuildPath` or `str`. The former two just create a
        `SourcePath` with the path within the existing object, whereas `str` is interpreted relative
        to `Env.cur_dir`.

        Parameters
        ----------
        :param env: the Environment
        :param path: the path object

        Returns
        -------
        A `SourcePath` object
        """

        if isinstance(path, SourcePath):
            return SourcePath(path._path)
        if isinstance(path, BuildPath):
            return SourcePath(path._path)
        return SourcePath(env.cur_dir + '/' + path)

    def __str__(self) -> str:
        return self._path

    def __repr__(self) -> str:
        return repr(self._path)


class BuildPath(str):
    """
    A path for build files

    This class is used to refer to build files. These are specified relative to the current
    directory and the build directory in the given environment (`Env.cur_dir`). So, basically build
    files are put into `"$NPBUILD/" + Env.cur_dir`.

    `BuildPath` implements `__str__` and `__repr__` and can thus be used as a string (the contained
    path).
    """

    def __init__(self, path: str):
        """
        Creates a new `BuildPath` that simply uses the given path

        Parameters
        ----------
        :param path: the path
        """

        self._path = path

    @staticmethod
    def new(env, path):
        """
        Creates a new `BuildPath` from given path object

        The path object can be a `SourcePath`, `BuildPath` or `str`. If it's build path, it simply
        creates a copy. If it's a source path, it produces a `BuildPath` consisting of
        `Env.build_dir` and the path. If it's a string, it produces a `BuildPath` consisting of
        `Env.build_dir`, `Env.cur_dir` and the string.

        Parameters
        ----------
        :param env: the Environment
        :param path: the path object

        Returns
        -------
        A `BuildPath` object
        """

        if isinstance(path, BuildPath):
            return BuildPath(path._path)
        if isinstance(path, SourcePath):
            return BuildPath(env.build_dir + '/' + path._path)
        return BuildPath(env.build_dir + '/' + env.cur_dir + '/' + path)

    @staticmethod
    def with_file_ext(env, path, ext: str):
        """
        Creates a new `BuildPath` from given path object and file extension

        This method is the same as `BuildPath.new`, but replaces the file extension of the given
        path with `ext`.

        Parameters
        ----------
        :param env: the Environment
        :param path: the path object
        :param ext: the new file extension

        Returns
        -------
        A `BuildPath` object
        """

        (root, _cur_ext) = os.path.splitext(path)
        if isinstance(path, BuildPath):
            return BuildPath(root + '.' + ext)
        if isinstance(path, SourcePath):
            return BuildPath.new(env, SourcePath(root + '.' + ext))
        return BuildPath.new(env, root + '.' + ext)

    def __str__(self) -> str:
        return self._path

    def __repr__(self) -> str:
        return repr(self._path)



import argparse
from glob import glob
import os
import subprocess
import sys

# read __version__ from relative from file to avoid that we need the ninjapie package
with open(os.path.dirname(__file__) + '/version.py', encoding='utf-8') as vfile:
    exec(vfile.read())  # pylint: disable=W0122


def all_files(build_dir):
    """Collects a string with all files found by the patters in the globs file"""

    files = ''
    try:
        with open(build_dir + '/.build.globs', 'r', encoding='utf-8') as file:
            for line in file.readlines():
                files += '\n'.join(glob(line.strip(), recursive=True))
                files += '\n'
    except FileNotFoundError:
        pass
    return files


def clean(build_dir, _args, _ninja_args):
    """Implements the clean command"""

    def remove_dir(path):
        """Removes the given directory recursively"""
        for entry in os.listdir(path):
            try:
                remove_dir(path + '/' + entry)
            except OSError:
                os.remove(path + '/' + entry)
        os.rmdir(path)

    remove_dir(build_dir)
    return 0


def build(build_dir, args, ninja_args):
    """Implements the build command"""

    reconf = args.force_regen
    root_dir = os.path.dirname(os.path.dirname(os.path.realpath(__file__)))
    all_files_path = build_dir + '/.build.files'
    build_file = build_dir + '/build.ninja'

    # export PYTHONPATH to find the ninjapie modules
    python_path = os.environ.setdefault('PYTHONPATH', '')
    os.environ['PYTHONPATH'] = root_dir + ': ' + python_path

    # check if we need to reconfigure
    if not reconf:
        try:
            # check whether files have been added or removed
            with open(all_files_path, 'r', encoding='utf-8') as file:
                old_files = file.read()
            new_files = all_files(build_dir)
            # if the list of files changed, we need to reconfigure
            reconf = old_files != new_files
        except FileNotFoundError:
            reconf = True

    # run configure if not done before or it's required
    if reconf or not os.path.isfile(build_file):
        try:
            # run build.py, but don't write *.pyc files
            subprocess.check_call(['python3', '-B', 'build.py'])
        except KeyboardInterrupt:
            return 1
        except Exception as exc:  # pylint: disable=W0718
            print("Executing build.py failed:", exc)
            return 1

        # store new list of files from globs
        new_files = all_files(build_dir)
        with open(all_files_path, 'w', encoding='utf-8') as file:
            file.write(new_files)

    # now build everything with ninja
    try:
        subprocess.check_call(['ninja', '-f', build_file] + ninja_args, stdout=sys.stderr.buffer)
    except KeyboardInterrupt:
        # ensure that we regenerate the build.ninja next time. Since ninja does not accept the
        # build.ninja, it will also not detect changes our build files in order to regenerate it.
        # Therefore, force a regenerate next time by removing the file.
        os.remove(all_files_path)
        return 1
    except Exception as exc:  # pylint: disable=W0718
        print("Running ninja failed:", exc)
        os.remove(all_files_path)
        return 1

    return 0


def main(argv=None):
    """Main entry point"""

    if argv is None:
        argv = sys.argv[1:]

    # determine and create build dir
    build_dir = os.environ.setdefault('NPBUILD', 'build')
    if not os.path.isdir(build_dir):
        os.makedirs(build_dir)

    # define command line arguments
    parser = argparse.ArgumentParser(
        description='This is the Ninjapie build system. '
        'Besides the supported command line arguments all additional arguments, '
        'preceeded by "--", will be passed to ninja. For example "ninjapie -- -v".'
    )
    version = __version__  # pylint: disable=E0602
    parser.add_argument('--version', action='version',
                        version='%(prog)s {version}'.format(version=version))

    subparsers = parser.add_subparsers(
        title='commands',
        description='The command to execute (build by default)',
    )

    parser_clean = subparsers.add_parser('clean', description='Removes the build directory')
    parser_clean.set_defaults(func=clean)

    parser_build = subparsers.add_parser(
        'build', description='Builds everything (the default command)')
    parser_build.add_argument('-f', '--force-regen', action='store_true', default=False,
                              help='force a regeneration of the ninja build file')
    parser_build.set_defaults(func=build)

    # we pass everything after "--" directly to ninja
    try:
        ninja_start = argv.index('--')
        ninja_args = argv[ninja_start + 1:]
        our_args = argv[0:ninja_start]
    except ValueError:
        ninja_args = []
        our_args = argv

    # parse arguments and run command
    args = parser.parse_args(our_args)
    try:
        res = args.func(build_dir, args, ninja_args)
    except AttributeError:
        # if func attribute is not found, fall back to default: build
        args.force_regen = False
        res = build(build_dir, args, ninja_args)
    return res


if __name__ == "__main__":
    sys.exit(main())



import copy
from glob import glob
import importlib.util
from pathlib import Path
import os
import sys

from ninjapie.path import BuildPath, SourcePath
from ninjapie.generator import BuildEdge, Generator


class Env:
    """
    An `Env` is a container for variables that are used to produce build edges for the `Generator`.

    The variables are tools (e.g., `CXX`), compiler flags (e.g., `CFLAGS`), and paths (e.g.,
    `CPPPATH`). These variables can be changed and, when their values are as desired, build edges
    can be produced based on them. In particular, `Env` can be cloned to change a variable for a
    particular build edge without influencing anything else.

    Furthermore, `Env` holds a current working directory (`Env.cur_dir`) which is used to specify
    the paths to input files relatively. The output files are put into the build directory, given by
    the environment variable `$NPBUILD`.

    An example usage of `Env` looks like the following:
    >>> env = Env()
    ... gen = Generator()
    ... env['CXXFLAGS'] += ['-Wall', '-Wextra']
    ... env.cxx_exe(gen, out='hello', ins=['hello.cc'])
    ... gen.write_to_file()
    """

    class _Location:
        def __init__(self, path: str):
            self.path = path

    def __init__(self):
        """
        Creates a new `Env` with default settings.

        By default, the tools are set to the gcc and binutils toolchain (`CXX`=`g++`, `AR`=`gcc-ar`,
        etc.), compiler flags are empty, and paths are empty as well.
        """

        self._id = 1
        self._cwd = Env._Location('.')
        self._build_dir = os.environ.get('NPBUILD')
        self._vars = {}

        # default tools
        self._vars['CXX'] = 'g++'
        self._vars['CPP'] = 'cpp'
        self._vars['AS'] = 'gcc'
        self._vars['CC'] = 'gcc'
        self._vars['AR'] = 'gcc-ar'
        self._vars['SHLINK'] = 'gcc'
        self._vars['RANLIB'] = 'gcc-ranlib'
        self._vars['STRIP'] = 'strip'
        self._vars['CARGO'] = 'cargo'

        # default flags
        self._vars['ASFLAGS'] = []
        self._vars['CFLAGS'] = []
        self._vars['CPPFLAGS'] = []
        self._vars['CXXFLAGS'] = []
        self._vars['LINKFLAGS'] = []
        self._vars['SHLINKFLAGS'] = []
        self._vars['ARFLAGS'] = []
        self._vars['INSTFLAGS'] = []

        # default paths
        self._vars['CPPPATH'] = []
        self._vars['LIBPATH'] = []

        # default rust settings
        self._vars['RUSTOUT'] = '.'
        self._vars['CRGFLAGS'] = []
        self._vars['CRGENV'] = {}

    def clone(self):
        """
        Clones this environment to produce an independently changable copy.

        The clone can be used to change the variables of the environment in order to prepare for a
        build edge with different settings without influencing other build edges.

        An example usage looks like the following:
        >>> env['CFLAGS'] += ['-Wall', '-Wextra']
        ...
        ... foo_env = env.clone()
        ... foo_env.add_flag('CPPFLAGS', '-DMY_CONSTANT=42')
        ... foo_env.remove_flag('CFLAGS', '-Wextra')
        ... obj = foo_env.cc(gen, out='foo.o', ins=['foo.c'])
        ...
        ... env.c_exe(gen, out='hello', ins=['hello.c', obj])

        The example has a general environment with default settings and clones this environment to
        produce an object file that requires different settings to be built. Afterwards, the
        original environment is used to link the application.
        """

        env = type(self)()
        env._id = self._id + 1
        env._cwd = self._cwd
        env._vars = copy.deepcopy(self._vars)
        return env

    @property
    def cur_dir(self) -> str:
        """
        Returns the current directory where input files are expected
        """

        return self._cwd.path

    @property
    def build_dir(self) -> str:
        """
        Returns the build directory where all output files are placed (potentially in sub
        directories)
        """

        return self._build_dir

    def __getitem__(self, var: str):
        """
        Returns the value or the variable with given name

        Parameters
        ----------
        :param var: the variable name

        Returns
        -------
        The value of the variable
        """

        return self._vars[var]

    def __setitem__(self, var: str, value):
        """
        Sets the value or the variable with given name to `value`

        Parameters
        ----------
        :param var: the variable name
        :param value: the new value
        """

        self._vars[var] = value

    def add_flag(self, var: str, flag: str):
        """
        Adds a flag to the given flag variable

        This is a convenience method that is semantically equivalent to:
        >>> self[var] += [flag]

        Note that this can only be used for variables of type `list[str]` such as `CFLAGS`.

        Parameters
        ----------
        :param var: the variable name
        :param flag: the flag to add
        """

        self.add_flags(var, [flag])

    def add_flags(self, var: str, flags: list[str]):
        """
        Adds flags to the given flag variable

        This is a convenience method that is semantically equivalent to:
        >>> self[var] += flags

        Note that this can only be used for variables of type `list[str]` such as `CFLAGS`.

        Parameters
        ----------
        :param var: the variable name
        :param flags: a list of flags to add
        """

        assert isinstance(self._vars[var], list)
        self._vars[var] += flags

    def remove_flag(self, var: str, flag: str):
        """
        Removes the given flag from the given flag variable

        This is a convenience method that is semantically equivalent to:
        >>> self[var].remove(flag)

        Note that this can only be used for variables of type `list[str]` such as `CFLAGS`.

        Parameters
        ----------
        :param var: the variable name
        :param flag: the flag to remove
        """

        self.remove_flags(var, [flag])

    def remove_flags(self, var: str, flags: list[str]):
        """
        Removes the given flags from the given flag variable

        This is a convenience method that is semantically equivalent to:
        >>> for flag in flags:
        ...     self[var].remove(flag)

        Note that this can only be used for variables of type `list[str]` such as `CFLAGS`.

        Parameters
        ----------
        :param var: the variable name
        :param flags: the flags to remove
        """

        for flag in flags:
            assert isinstance(self._vars[var], list)
            if flag in self._vars[var]:
                self._vars[var].remove(flag)

    def sub_build(self, gen: Generator, dir: str):
        """
        Calls the build.py in the given subdirectory

        In a project with multiple files to build (libraries, executables, etc.), it makes sense to
        organize them into multiple directories. This method is used to run the `build.py` in a
        subdirectory and thus add the rules and build edges produced there to the generator. Before
        running the `build.py`, the current directory (`Env.cur_dir`) will be changed to the
        subdirectory. This allows the `build.py` in the subdirectory to use paths relative to its
        own directory. Of course, `Env.sub_build` can be nested arbitrarily deep.

        The `build.py` files in the subdirectory are expected to contain a function `build` with the
        generator and environment as arguments. This will be called by `Env.sub_build`.

        For example, the `build.py` in the root directory could do the following:
        >>> env.sub_build(gen, 'sub')
        ... env.c_exe(gen, out='hello', ins=['hello.c'], libs=['sub'])
        ... gen.write_to_file()
        And the `build.py` in the subdirectory `sub`:
        >>> def build(gen, env):
        ...     env.static_lib(gen, out='sub', ins=['sub.c'])

        This example would produce `libsub.a` in the subdirectory and use it to produce the `hello`
        executable in the root directory.

        Parameters
        ----------
        :param gen: the generator
        :param dir: the subdirectory to enter

        Returns
        -------
        The return value of the build function in the subdirectory
        """

        old_cwd = self.cur_dir
        self._cwd.path += '/' + dir

        gen._add_build_file(self.cur_dir + '/build.py')

        # import module
        mod_path = self.cur_dir[2:].replace('/', '.')
        spec = importlib.util.spec_from_file_location(mod_path, self.cur_dir + '/build.py')
        sub = importlib.util.module_from_spec(spec)
        sys.modules[spec.name] = sub
        spec.loader.exec_module(sub)

        # call build function in module
        res = sub.build(gen, self)

        self._cwd.path = old_cwd
        return res

    def glob(self, gen: Generator, pattern: str) -> list[SourcePath]:
        """
        Produces a list of files that match the given pattern

        The pattern uses the same syntax as the Python function `glob.glob`. For example, '*.c'
        produces a list of all C files in the current directory. Since the `recursive` argument to
        `glob.glob` is always set to true, '**/*.c' would produce a list of all C files in the
        current directory or subdirectories.

        Note that globbing has the side effect that the required build steps might change on added
        or removed files. For that reason, Ninjapie records the glob patterns and regenerates the
        ninja build file whenever any file is added or removed. In other words, using `Env.glob` is
        a trade-off between more convenience and faster builds, because Ninjapie needs to perform
        additional checks for globs. Note that this also means that *only* this function should be
        used for globbing, because all other ways bypass Ninjapie and therefore lead to potentially
        outdated ninja build files.

        Parameters
        ----------
        :param gen: the generator
        :param pattern: the pattern for globbing

        Returns
        -------
        The list of found files as `SourcePath` objects
        """

        pat = SourcePath.new(self, pattern)
        gen._add_glob(pat)
        files = glob(pat, recursive=True)
        return [SourcePath(f) for f in files]

    def install(self, gen: Generator, outdir: str, input: str) -> str:
        """
        Installs the given input file into the given output directory

        Parameters
        ----------
        :param gen: the generator
        :param outdir: the output directory
        :param input: the input file

        Returns
        -------
        The path of the installed file
        """

        return self.install_as(gen, outdir + '/' + os.path.basename(input), input)

    def install_as(self, gen: Generator, out: str, input: str) -> str:
        """
        Installs the given input file into the given output file

        Parameters
        ----------
        :param gen: the generator
        :param out: the output file
        :param input: the input file

        Variables
        ---------
        :param `INSTFLAGS`: the flags (e.g., ['--mode=0644'])

        Returns
        -------
        The path of the installed file
        """

        flags = ' '.join(self['INSTFLAGS'])
        edge = BuildEdge(
            'install',
            outs=[out],
            ins=[SourcePath.new(self, input)],
            vars={
                'instflags': flags
            }
        )
        gen.add_build(edge)
        return out

    def strip(self, gen: Generator, out: str, input: str) -> BuildPath:
        """
        Strips the symbols from the given input file

        Parameters
        ----------
        :param gen: the generator
        :param out: the output file
        :param input: the input file

        Variables
        ---------
        :param `STRIP`: the tool name (e.g., 'arm-none-eabi-strip')

        Returns
        -------

        A `BuildPath` to the stripped file
        """

        bin = BuildPath.new(self, out)
        edge = BuildEdge(
            'strip',
            outs=[bin],
            ins=[SourcePath.new(self, input)],
            vars={
                'strip': self['STRIP']
            }
        )
        gen.add_build(edge)
        return bin

    def cpp(self, gen: Generator, out: str, input: str) -> BuildPath:
        """
        Runs the C preprocessor on the given input file

        Parameters
        ----------
        :param gen: the generator
        :param out: the output file
        :param input: the input file

        Variables
        ---------
        :param `CPP`: the tool name (e.g., 'arm-none-eabi-cpp')
        :param `CPPFLAGS`: the flags (e.g., ['-DFOO=1'])
        :param `CPPPATH`: the include paths (e.g., ['include'])

        Returns
        -------
        A `BuildPath` to the output file
        """

        flags = ' '.join(self['CPPFLAGS'])
        flags += ' ' + ' '.join(['-I' + i for i in self['CPPPATH']])

        bin = BuildPath.new(self, out)
        edge = BuildEdge(
            'cpp',
            outs=[bin],
            ins=[SourcePath.new(self, input)],
            vars={
                'cpp': self['CPP'],
                'cppflags': flags
            }
        )
        gen.add_build(edge)
        return bin

    def asm(self, gen: Generator, out: str, ins: list[str]) -> BuildPath:
        """
        Runs the assember on the given input files

        Note that this method assumes that the C compiler (like gcc) can be used for this purpose.

        Parameters
        ----------
        :param gen: the generator
        :param out: the output file
        :param ins: the list of input files

        Variables
        ---------
        :param `CC`: the tool name (e.g., 'gcc')
        :param `ASFLAGS`: the flags (e.g., ['-Wa,--32'])
        :param `CPPFLAGS`: the preprocessor flags (e.g., ['-DFOO=1'])
        :param `CPPPATH`: the include paths (e.g., ['include'])

        Returns
        -------
        A `BuildPath` to the output file
        """

        flags = ' '.join(self['ASFLAGS'] + self['CPPFLAGS'])
        flags += ' ' + ' '.join(['-I' + i for i in self['CPPPATH']])
        return self._cc(gen, out, ins, flags)

    def cc(self, gen: Generator, out: str, ins: list[str]) -> BuildPath:  # pylint: disable=C0103
        """
        Runs the C compiler on the given input files

        Parameters
        ----------
        :param gen: the generator
        :param out: the output file
        :param ins: the list of input files

        Variables
        ---------
        :param `CC`: the tool name (e.g., 'gcc')
        :param `CFLAGS`: the flags (e.g., ['-Wall'])
        :param `CPPFLAGS`: the preprocessor flags (e.g., ['-DFOO=1'])
        :param `CPPPATH`: the include paths (e.g., ['include'])

        Returns
        -------
        A `BuildPath` to the output file
        """

        flags = ' '.join(self['CFLAGS'] + self['CPPFLAGS'])
        flags += ' ' + ' '.join(['-I' + i for i in self['CPPPATH']])
        return self._cc(gen, out, ins, flags)

    def _cc(self, gen: Generator, out: str, ins: list[str], flags: str) -> BuildPath:
        obj = BuildPath.new(self, out)
        edge = BuildEdge(
            'cc',
            outs=[obj],
            ins=[SourcePath.new(self, i) for i in ins],
            vars={
                'cc': self['CC'],
                'ccflags': flags
            }
        )
        gen.add_build(edge)
        return obj

    def cxx(self, gen: Generator, out: str, ins: list[str]) -> BuildPath:
        """
        Runs the C++ compiler on the given input files

        Parameters
        ----------
        :param gen: the generator
        :param out: the output file
        :param ins: the list of input files

        Variables
        ---------
        :param `CXX`: the tool name (e.g., 'g++')
        :param `CXXFLAGS`: the flags (e.g., ['-Wall'])
        :param `CPPFLAGS`: the preprocessor flags (e.g., ['-DFOO=1'])
        :param `CPPPATH`: the include paths (e.g., ['include'])

        Returns
        -------
        A `BuildPath` to the output file
        """

        flags = ' '.join(self['CXXFLAGS'] + self['CPPFLAGS'])
        flags += ' ' + ' '.join(['-I' + i for i in self['CPPPATH']])

        obj = BuildPath.new(self, out)
        edge = BuildEdge(
            'cxx',
            outs=[obj],
            ins=[SourcePath.new(self, i) for i in ins],
            vars={
                'cxx': self['CXX'],
                'cxxflags': flags
            }
        )
        gen.add_build(edge)
        return obj

    def objs(self, gen: Generator, ins: list[str]) -> list[BuildPath]:
        """
        Produces object files for the given input files

        This method will call `Env.asm`, `Env.cc`, or `Env.cxx` to build the input file, depending
        on the file extension. Object files or libraries are simply appended to list of output
        files.

        Parameters
        ----------
        :param gen: the generator
        :param ins: the list of input files

        Returns
        -------
        A list of `BuildPath`s to the object files
        """

        # add a per-environment suffix to allow users to build the same files in different
        # environments without interference
        suffix = str(self._id) + '.o'
        objs = []
        for i in ins:
            if i.endswith('.S') or i.endswith('.s'):
                objs.append(self.asm(gen, BuildPath.with_file_ext(self, i, suffix), [i]))
            elif i.endswith('.c'):
                objs.append(self.cc(gen, BuildPath.with_file_ext(self, i, suffix), [i]))
            elif i.endswith('.cc') or i.endswith('.cpp'):
                objs.append(self.cxx(gen, BuildPath.with_file_ext(self, i, suffix), [i]))
            elif i.endswith('.o') or i.endswith('.a') or i.endswith('.so'):
                objs.append(BuildPath.new(self, i))
        return objs

    def static_lib(self, gen: Generator, out: str, ins: list[str]) -> BuildPath:
        """
        Produces the static library `"lib" + out + ".a"` from given input files

        The input files can be all file types that `Env.objs` supports.

        Parameters
        ----------
        :param gen: the generator
        :param out: the output file
        :param ins: the list of input files

        Variables
        ---------
        :param `AR`: the tool name for archiving (e.g., 'gcc-ar')
        :param `RANLIB`: the tool name for index generation (e.g., 'gcc-ranlib')
        :param `ARFLAGS`: the flags

        Returns
        -------
        The `BuildPath` to the produced static library
        """

        flags = ' '.join(self['ARFLAGS'])
        lib = BuildPath.new(self, 'lib' + out + '.a')
        edge = BuildEdge(
            'ar',
            outs=[lib],
            ins=self.objs(gen, ins),
            vars={
                'ar': self['AR'],
                'ranlib': self['RANLIB'],
                'arflags': flags
            }
        )
        gen.add_build(edge)
        return lib

    def shared_lib(self, gen: Generator, out: str, ins: list[str]) -> BuildPath:
        """
        Produces the shared library `"lib" + out + ".so"` from given input files

        The input files can be all file types that `Env.objs` supports.

        Parameters
        ----------
        :param gen: the generator
        :param out: the output file
        :param ins: the list of input files

        Variables
        ---------
        :param `SHLINK`: the tool name to create the shared library (e.g., 'gcc')
        :param `SHLINKFLAGS`: the flags (e.g., ['-march=rv64imafdc'])

        Returns
        -------
        The `BuildPath` to the produced shared library
        """

        flags = ' '.join(self['SHLINKFLAGS'])
        lib = BuildPath.new(self, 'lib' + out + '.so')
        edge = BuildEdge(
            'shlink',
            outs=[lib],
            ins=self.objs(gen, ins),
            vars={
                'shlink': self['SHLINK'],
                'shlinkflags': flags
            }
        )
        gen.add_build(edge)
        return lib

    # pylint: disable=R0917
    def c_exe(self, gen: Generator, out: str, ins: list[str],
              libs: list[str] = None, deps: list[str] = None) -> BuildPath:
        """
        Produces a C executable from given input files

        The input files can be all file types that `Env.objs` supports.

        Parameters
        ----------
        :param gen: the generator
        :param out: the output file
        :param ins: the list of input files
        :param libs: the list of library names that the executable should be linked against
        :param deps: the additional list of dependencies

        Variables
        ---------
        :param `CC`: the tool name (e.g., 'gcc')
        :param `LINKFLAGS`: the flags (e.g., ['-march=rv64imafdc'])
        :param `LIBPATH`: the paths to search libraries in (e.g., ['lib'])

        Returns
        -------
        The `BuildPath` to the produced executable
        """

        libs = [] if libs is None else libs
        deps = [] if deps is None else deps
        return self._c_cxx_exe(gen, out, ins, libs, deps, self['CC'])

    # pylint: disable=R0917
    def cxx_exe(self, gen: Generator, out: str, ins: list[str],
                libs: list[str] = None, deps: list[str] = None) -> BuildPath:
        """
        Produces a C++ executable from given input files

        The input files can be all file types that `Env.objs` supports.

        Parameters
        ----------
        :param gen: the generator
        :param out: the output file
        :param ins: the list of input files
        :param libs: the list of library names that the executable should be linked against
        :param deps: the additional list of dependencies

        Variables
        ---------
        :param `CXX`: the tool name (e.g., 'g++')
        :param `LINKFLAGS`: the flags (e.g., ['-march=rv64imafdc'])
        :param `LIBPATH`: the paths to search libraries in (e.g., ['lib'])

        Returns
        -------
        The `BuildPath` to the produced executable
        """

        libs = [] if libs is None else libs
        deps = [] if deps is None else deps
        return self._c_cxx_exe(gen, out, ins, libs, deps, self['CXX'])

    # pylint: disable=R0917
    def _c_cxx_exe(self, gen: Generator, out: str, ins: list[str],
                   libs: list[str], deps: list[str], linker: str) -> BuildPath:
        flags = ' '.join(self['LINKFLAGS'])
        if len(libs) > 0:
            flags += ' ' + ' '.join(['-L' + dir for dir in self['LIBPATH']])
            flags += ' -Wl,--start-group'
            flags += ' ' + ' '.join(['-l' + lib for lib in libs])
            flags += ' -Wl,--end-group'

        bin = BuildPath.new(self, out)
        edge = BuildEdge(
            'link',
            outs=[bin],
            ins=self.objs(gen, ins),
            deps=deps,
            libs=libs,
            lib_path=self['LIBPATH'],
            vars={
                'link': linker,
                'linkflags': flags
            }
        )
        gen.add_build(edge)
        return bin

    def rust_lib(self, gen: Generator, out: str, deps: list[str] = None) -> BuildPath:
        """
        Produces a Rust library

        This method runs `cargo` in the current directory and therefore expects a `Cargo.toml` that
        produces a static Rust library. The `--target-dir` argument will be pass to cargo according
        to `Env.build_dir` and `RUSTOUT`. The location of the produced file will be determined by
        the `--target` and `--release` arguments in `CRGFLAGS`, if present. You can also use
        `CRGENV` to provide additional environment variables to cargo.

        Note: if `deps` is empty, this build edge will always be rebuilt to let `cargo` determine
        the required actions.

        Parameters
        ----------
        :param gen: the generator
        :param out: the output file
        :param deps: the additional list of dependencies

        Variables
        ---------
        :param `CRGFLAGS`: the flags (e.g., ['--release'])
        :param `RUSTOUT`: an optional subdirectory in `Env.build_dir` for Rust outputs
        :param `CRGENV`: additional environment variables

        Returns
        -------
        The `BuildPath` to the produced static library
        """

        deps = [] if deps is None else deps
        return self.rust(gen, ['lib' + out + '.a'], deps)[0]

    def rust_exe(self, gen: Generator, out: str, deps: list[str] = None) -> BuildPath:
        """
        Produces a Rust executable

        This method runs `cargo` in the current directory and therefore expects a `Cargo.toml` that
        produces a binary. The `--target-dir` argument will be pass to cargo according to
        `Env.build_dir` and `RUSTOUT`. The location of the produced file will be determined by the
        `--target` and `--release` arguments in `CRGFLAGS`, if present. You can also use `CRGENV` to
        provide additional environment variables to cargo.

        Note: if `deps` is empty, this build edge will always be rebuilt to let `cargo` determine
        the required actions.

        Parameters
        ----------
        :param gen: the generator
        :param out: the output file
        :param deps: the additional list of dependencies

        Variables
        ---------
        :param `CRGFLAGS`: the flags (e.g., ['--release'])
        :param `RUSTOUT`: an optional subdirectory in `Env.build_dir` for Rust outputs
        :param `CRGENV`: additional environment variables

        Returns
        -------
        The `BuildPath` to the produced executable
        """

        deps = [] if deps is None else deps
        return self.rust(gen, [out], deps)[0]

    def rust(self, gen: Generator, outs: list[str], deps: list[str]) -> [BuildPath]:
        """
        Produces multiple Rust libraries or executables

        This method runs `cargo` in the current directory and therefore expects a `Cargo.toml`. The
        `--target-dir` argument will be pass to cargo according to `Env.build_dir` and `RUSTOUT`.
        The location of the produced files will be determined by the `--target` and `--release`
        arguments in `CRGFLAGS`, if present. You can also use `CRGENV` to provide additional
        environment variables to cargo.

        Note: if `deps` is empty, this build edge will always be rebuilt to let `cargo` determine
        the required actions.

        Parameters
        ----------
        :param gen: the generator
        :param outs: the output files
        :param deps: the additional list of dependencies

        Variables
        ---------
        :param `CRGFLAGS`: the flags (e.g., ['--release'])
        :param `RUSTOUT`: an optional subdirectory in `Env.build_dir` for Rust outputs
        :param `CRGENV`: additional environment variables

        Returns
        -------
        A list of `BuildPath`s to the produced files
        """

        # determine whether cargo puts the output in a target-specific directory
        target_dir = ''
        try:
            idx = self['CRGFLAGS'].index('--target')
            # if it's a path to the spec, the triple is the filename without extension
            if self['CRGFLAGS'][idx + 1].endswith('.json'):
                target_dir = Path(self['CRGFLAGS'][idx + 1]).stem
            # otherwise it's already the triple we need
            else:
                target_dir = self['CRGFLAGS'][idx + 1]
            target_dir += '/'
        except ValueError:
            pass

        # determine destination directory
        btype = 'release' if '--release' in self['CRGFLAGS'] else 'debug'
        dest_dir = BuildPath(self.build_dir + '/' + self['RUSTOUT'] + '/' + target_dir + btype)
        out_paths = [BuildPath(dest_dir + '/' + o) for o in outs]

        flags = ' '.join(self['CRGFLAGS'])
        # make sure that cargo puts it there
        flags += ' --target-dir "' + os.path.abspath(self.build_dir + '/' + self['RUSTOUT']) + '"'

        # build environment variables
        vars_str = ''
        for key, value in self['CRGENV'].items():
            vars_str += ' ' + key + '="' + value + '"'

        edge = BuildEdge(
            'cargo',
            outs=out_paths,
            ins=[],
            deps=deps,
            vars={
                'cargo': self['CARGO'],
                'dir': self.cur_dir,
                'cargoflags': 'build ' + flags,
                'env': vars_str
            }
        )
        gen.add_build(edge)
        return out_paths
